<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE ncx PUBLIC "-//NISO//DTD ncx 2005-1//EN"
	"http://www.daisy.org/z3986/2005/ncx-2005-1.dtd">

<ncx xmlns="http://www.daisy.org/z3986/2005/ncx/" version="2005-1" xml:lang="en-US">
<head>
  <meta name="dtb:uid" content="RailsGuides"/>
  <meta name="dtb:depth" content="2"/>
  <meta name="dtb:totalPageCount" content="0"/>
  <meta name="dtb:maxPageNumber" content="0"/>
</head>
<docTitle><text>Ruby on Rails Guides</text></docTitle>
<docAuthor><text>docrails</text></docAuthor>
<navMap>
  <navPoint playOrder="0" class="periodical" id="periodical">
    <navLabel>
      <text>Table of Contents</text>
    </navLabel>
    <content src="toc.html"/>

    <navPoint class="section" id="welcome" playOrder="1">
      <navLabel>
        <text>Introduction</text>
      </navLabel>
      <content src="welcome.html"/>

      <navPoint class="article" id="welcome" playOrder="2">
        <navLabel>
          <text>Welcome</text>
        </navLabel>
        <content src="welcome.html"/>
      </navPoint>
      <navPoint class="article" id="copyright" playOrder="4">
        <navLabel><text>Copyright &amp; License</text></navLabel>
        <content src="copyright.html"/>
      </navPoint>
    </navPoint>

    <% play_order = 4 %>
    <% documents_by_section.each_with_index do |section, section_no| %>
      <navPoint class="section" id="chapter_<%= section_no + 1 %>" playOrder="<% play_order +=1 %>">
        <navLabel>
          <text><%= section['name'] %></text>
        </navLabel>
        <content src="<%=section['documents'].first['url'] %>"/>

        <% section['documents'].each_with_index do |document, document_no| %>
          <navPoint class="article" id="_<%=section_no+1%>.<%=document_no+1%>" playOrder="<%=play_order +=1 %>">
            <navLabel>
              <text><%= document['name'] %></text>
            </navLabel>
            <content src="<%=document['url'] %>"/>
          </navPoint>
        <% end %>
      </navPoint>
    <% end %>

  </navPoint>
</navMap>
</ncx>
