<header>
  <h1>Blocked host: <%= @host %></h1>
</header>
<div id="container">
  <h2>To allow requests to <%= @host %> make sure it is a valid hostname (containing only numbers, letters, dashes and dots), then add the following to your environment configuration:</h2>
  <pre>config.hosts &lt;&lt; "<%= @host %>"</pre>
</div>
