Rails.root: <%= defined?(Rails) && Rails.respond_to?(:root) ? Rails.root : "unset" %>

<% @traces.each do |name, trace| %>
<% if trace.any? %>
<%= name %>
<%= trace.map { |t| t[:trace] }.join("\n") %>

<% end %>
<% end %>
