# frozen_string_literal: true

TEST_ROOT       = __dir__
ASSETS_ROOT     = TEST_ROOT + "/assets"
FIXTURES_ROOT   = TEST_ROOT + "/fixtures"
MIGRATIONS_ROOT = TEST_ROOT + "/migrations"
SCHEMA_ROOT     = TEST_ROOT + "/schema"
