<%= render 'welcome' %>

<h3>Kindle Edition</h3>

<div>
  The Kindle Edition of the Rails Guides should be considered a work in progress. Feedback is really welcome. Please see the "Feedback" section at the end of each guide for instructions.
</div>
