<?xml version="1.0" encoding="utf-8"?>

<package xmlns="http://www.idpf.org/2007/opf" version="2.0" unique-identifier="RailsGuides">
<metadata>
  <meta name="cover" content="cover" />
  <dc-metadata xmlns:dc="http://purl.org/dc/elements/1.1/">

       <dc:title>Ruby on Rails Guides (<%= @version || "main@#{@edge[0, 7]}" %>)</dc:title>

  	<dc:language>en-us</dc:language>
  	<dc:creator>Ruby on Rails</dc:creator>
  	<dc:publisher>Ruby on Rails</dc:publisher>
  	<dc:subject>Reference</dc:subject>
  	<dc:date><%= Time.now.strftime('%Y-%m-%d') %></dc:date>

    <dc:description>These guides are designed to make you immediately productive with Rails, and to help you understand how all of the pieces fit together.</dc:description>
  </dc-metadata>
  <x-metadata>
    <output content-type="application/x-mobipocket-subscription-magazine" encoding="utf-8"/>
  </x-metadata>
</metadata>

<manifest>
  <!-- HTML content files [mandatory] -->
  <% documents_flat.each do |document| %>
    <item id="<%= document['url'] %>" media-type="text/html" href="<%= document['url'] %>" />
  <% end %>

  <% %w{toc.html welcome.html copyright.html}.each do |url| %>
    <item id="<%= url %>" media-type="text/html" href="<%= url %>" />
  <% end %>

  <item id="toc" media-type="application/x-dtbncx+xml" href="toc.ncx" />

  <item id="cover" media-type="image/jpg" href="images/rails_guides_kindle_cover.jpg"/>
</manifest>

<spine toc="toc">
  <itemref idref="toc.html" />
  <itemref idref="welcome.html" />
  <itemref idref="copyright.html" />
  <% documents_flat.each do |document| %>
    <itemref idref="<%= document['url'] %>" />
  <% end %>
</spine>

<guide>
	<reference type="toc" title="Table of Contents" href="toc.html"></reference>
</guide>

</package>
