<% provide :title, "Deliver new inbound email by source" %>

<h1>Deliver new inbound email by source</h1>

<%= form_with(url: main_app.rails_conductor_inbound_email_sources_path, local: true) do |form| %>
  <div>
    <%= form.label :source, "Source" %><br>
    <%= form.text_area :source, size: "80x60" %>
  </div>

  <%= form.submit "Deliver inbound email" %>
<% end %>
