<h1>Adding an inline image while rendering</h1>

<% controller.attachments.inline["controller_attachments.jpg"] = 'via controller.attachments.inline' %>
<%= image_tag attachments['controller_attachments.jpg'].url %>

<% attachments.inline["attachments.jpg"] = 'via attachments.inline' %>
<%= image_tag attachments['attachments.jpg'].url %>
