/* Guides.rubyonrails.org */
/* Print.css */
/* Created January 30, 2009 */
/* Modified January 31, 2009
--------------------------------------- */

body, .wrapper, .note, .info, code, #topNav, .L, .R, #frame, #container, #header, #navigation, #footer, #feature, #mainCol, #subCol, .content {position: static; text-align: left; text-indent: 0; background: White; color: Black; border-color: Black; width: auto; height: auto; display: block; float: none; min-height: 0; margin: 0; padding: 0;}

body {
  background: #FFF;
  font-size: 10pt !important;
  font-family: "Helvetica Neue", Helvetica, Arial, sans-serif;
  line-height: 1.5;
  color: #000;
  padding: 0 3%;
  }

.hide, .nav {
  display: none !important;
  }

a:link, a:visited {
  background: transparent;
  font-weight: bold;
  text-decoration: underline;
  }

hr {
  background:#ccc;
  color:#ccc;
  width:100%;
  height:2px;
  margin:2em 0;
  padding:0;
  border:none;
}

h1,h2,h3,h4,h5,h6 { font-family: "Helvetica Neue", Arial, "Lucida Grande", sans-serif; }
code { font:.9em "Courier New", Monaco, Courier, monospace; display:inline}

img { float:left; margin:1.5em 1.5em 1.5em 0; }
a img { border:none; }

blockquote {
  margin:1.5em;
  padding:1em;
  font-style:italic;
  font-size:.9em;
}

.small  { font-size: .9em; }
.large  { font-size: 1.1em; }
