<!DOCTYPE html>
<html lang="en">
<head>
  <meta charset="utf-8" />
  <title><%= @page_title %></title>
  <style>
    body { background-color: #fff; color: #333; }

    body, p, ol, ul, td {
      font-family: helvetica, verdana, arial, sans-serif;
      font-size:   13px;
      line-height: 18px;
    }

    pre {
      background-color: #eee;
      padding: 10px;
      font-size: 11px;
      white-space: pre-wrap;
    }

    a { color: #000; }
    a:visited { color: #666; }
    a:hover { color: #fff; background-color:#000; }

    h2 { padding-left: 10px; }

    <%= yield :style %>
  </style>
</head>
<body>

<%= yield %>

</body>
</html>
