<% content_for :page_title, "Ruby on Rails Guides" %>
<% content_for :description, "Ruby on Rails Guides" %>

<% content_for :header_section do %>
<%= render 'welcome' %>
<% end %>

<% content_for :index_section do %>
<div id="subCol">
  <dl>
    <dt></dt>
    <% unless @edge -%>
      <dd class="kindle">Rails Guides are also available for <%= link_to 'Kindle', @mobi %>.</dd>
    <% end -%>
    <dd class="work-in-progress">Guides marked with this icon are currently being worked on and will not be available in the Guides Index menu. While still useful, they may contain incomplete information and even errors. You can help by reviewing them and posting your comments and corrections.</dd>
  </dl>
</div>
<% end %>

<% documents_by_section.each do |section| %>
  <h3><%= section['name'] %></h3>
  <dl>
    <% section['documents'].each do |document| %>
      <%= guide(document['name'], document['url'], work_in_progress: document['work_in_progress']) do %>
        <p><%= document['description'] %></p>
      <% end %>
    <% end %>
  </dl>
<% end %>
