<h2>Ruby on Rails Guides (<%= @edge ? @edge[0, 7] : @version %>)</h2>

<% if @edge %>
<p>
  These are <b>Edge Guides</b>, based on <a href="https://github.com/rails/rails/tree/<%= @edge %>">main@<%= @edge[0, 7] %></a>.
</p>
<p>
  If you are looking for the ones for the stable version, please check
  <a href="https://guides.rubyonrails.org">https://guides.rubyonrails.org</a> instead.
</p>
<% else %>
<p>
  These are the new guides for Rails 6.1 based on <a href="https://github.com/rails/rails/tree/<%= @version %>"><%= @version %></a>.
  These guides are designed to make you immediately productive with Rails, and to help you understand how all of the pieces fit together.
</p>
<% end %>
<p>
The guides for earlier releases:
<a href="https://guides.rubyonrails.org/v6.0/">Rails 6.0</a>,
<a href="https://guides.rubyonrails.org/v5.2/">Rails 5.2</a>,
<a href="https://guides.rubyonrails.org/v5.1/">Rails 5.1</a>,
<a href="https://guides.rubyonrails.org/v5.0/">Rails 5.0</a>,
<a href="https://guides.rubyonrails.org/v4.2/">Rails 4.2</a>,
<a href="https://guides.rubyonrails.org/v4.1/">Rails 4.1</a>,
<a href="https://guides.rubyonrails.org/v4.0/">Rails 4.0</a>,
<a href="https://guides.rubyonrails.org/v3.2/">Rails 3.2</a>,
<a href="https://guides.rubyonrails.org/v3.1/">Rails 3.1</a>,
<a href="https://guides.rubyonrails.org/v3.0/">Rails 3.0</a>, and
<a href="https://guides.rubyonrails.org/v2.3/">Rails 2.3</a>.
</p>
