<%# Template Dependency: messages/message %>

<%= render "header" %>
<%= render "comments/comments" %>

<%= render "messages/actions/move" %>

<%= render @message.history.events %>

<%# render "something_missing"   %>
<%# render "something_missing_1" %>

<%
  # Template Dependency: messages/form
%>