# frozen_string_literal: true

module ActiveStorage
  # Returns the version of the currently loaded \Rails as a +Gem::Version+
  def self.gem_version
    Gem::Version.new VERSION::STRING
  end

  module VERSION
    MAJOR = 1
    MINOR = 0
    TINY  = 0
    PRE   = "alpha"

    STRING = [MAJOR, MINOR, TINY, PRE].compact.join(".")
  end
end
