<!DOCTYPE html>
<html lang="en">
<head>
  <meta charset="utf-8">
  <meta name="viewport" content="width=device-width, initial-scale=1">
  <title><%= yield(:page_title) %></title>
  <link rel="stylesheet" type="text/css" href="stylesheets/style.css" data-turbolinks-track="reload">
  <link rel="stylesheet" type="text/css" href="stylesheets/print.css" media="print">
  <link rel="stylesheet" type="text/css" href="stylesheets/highlight.css" data-turbolinks-track="reload">
  <link href="images/favicon.ico" rel="shortcut icon" type="image/x-icon" />
  <script src="javascripts/turbolinks.js" data-turbolinks-track="reload"></script>
  <script src="javascripts/clipboard.js" data-turbolinks-track="reload"></script>
  <script src="javascripts/guides.js" data-turbolinks-track="reload"></script>
  <meta property="og:title" content="<%= yield(:page_title) %>" />
  <meta name="description" content="<%= yield(:description) %>" />
  <meta property="og:description" content="<%= yield(:description) %>" />
  <meta property="og:locale" content="en_US" />
  <meta property="og:site_name" content="Ruby on Rails Guides" />
  <meta property="og:image" content="https://avatars.githubusercontent.com/u/4223" />
  <meta property="og:type" content="website" />
</head>
<body class="guide">
  <% if badge_version = @edge ? "edge" : @version %>
  <div>
    <div id="version-badge"><%= badge_version %></div>
  </div>
  <% end %>
  <div id="topNav">
    <div class="wrapper">
      <strong class="more-info-label">More at <a href="https://rubyonrails.org/">rubyonrails.org:</a> </strong>
      <span class="red-button more-info-button">
        More Ruby on Rails
      </span>
      <ul class="more-info-links s-hidden">
        <li class="more-info"><a href="https://rubyonrails.org/blog">Blog</a></li>
        <li class="more-info"><a href="https://guides.rubyonrails.org/">Guides</a></li>
        <li class="more-info"><a href="https://api.rubyonrails.org/">API</a></li>
        <li class="more-info"><a href="https://discuss.rubyonrails.org/">Forum</a></li>
        <li class="more-info"><a href="https://github.com/rails/rails">Contribute on GitHub</a></li>
      </ul>
    </div>
  </div>
  <div id="header">
    <div class="wrapper clearfix">
      <h1><a href="index.html" title="Return to home page">Guides.rubyonrails.org</a></h1>
      <ul class="nav">
        <li><a class="nav-item" href="index.html">Home</a></li>
        <li class="guides-index guides-index-large">
          <a href="index.html" id="guidesMenu" class="guides-index-item nav-item">Guides Index</a>
          <div id="guides" class="clearfix" style="display: none;">
            <hr />
            <dl class="guides-section-container">
              <% documents_by_section.each do |section| %>
                <div class="guides-section">
                  <dt><%= section['name'] %></dt>
                  <% finished_documents(section['documents']).each do |document| %>
                  <dd><a href="<%= document['url'] %>"><%= document['name'] %></a></dd>
                  <% end %>
                </div>
              <% end %>
            </dl>
          </div>
        </li>
        <li><a class="nav-item" href="contributing_to_ruby_on_rails.html">Contribute</a></li>
        <li class="guides-index guides-index-small">
          <select class="guides-index-item nav-item">
            <option value="index.html">Guides Index</option>
            <% docs_for_menu.each do |section| %>
              <optgroup label="<%= section['name'] %>">
                <% finished_documents(section['documents']).each do |document| %>
                  <option value="<%= document['url'] %>"><%= document['name'] %></option>
                <% end %>
              </optgroup>
            <% end %>
          </select>
        </li>
      </ul>
    </div>
  </div>
  <hr class="hide" />

  <div id="feature">
    <div class="wrapper">
      <%= yield :header_section %>

      <%= yield :index_section %>
    </div>
  </div>

  <div id="container">
    <div class="wrapper">
      <div id="mainCol">
        <%= yield %>

        <h3>Feedback</h3>
        <p>
          You're encouraged to help improve the quality of this guide.
        </p>
        <p>
          Please contribute if you see any typos or factual errors.
          To get started, you can read our <%= link_to 'documentation contributions', 'https://edgeguides.rubyonrails.org/contributing_to_ruby_on_rails.html#contributing-to-the-rails-documentation' %> section.
        </p>
        <p>
          You may also find incomplete content or stuff that is not up to date.
          Please do add any missing documentation for main. Make sure to check
          <%= link_to 'Edge Guides', 'https://edgeguides.rubyonrails.org' %> first to verify
          if the issues are already fixed or not on the main branch.
          Check the <%= link_to 'Ruby on Rails Guides Guidelines', 'ruby_on_rails_guides_guidelines.html' %>
          for style and conventions.
        </p>
        <p>
          If for whatever reason you spot something to fix but cannot patch it yourself, please
          <%= link_to 'open an issue', 'https://github.com/rails/rails/issues' %>.
        </p>
        <p>And last but not least, any kind of discussion regarding Ruby on Rails
          documentation is very welcome on the <%= link_to 'rubyonrails-docs mailing list', 'https://discuss.rubyonrails.org/c/rubyonrails-docs' %>.
        </p>
      </div>
    </div>
  </div>

  <hr class="hide" />
  <div id="footer">
    <div class="wrapper">
      <%= render 'license' %>
    </div>
  </div>
</body>
</html>
