<p id="notice"><%= notice %></p>

<h1 id="subject"><%= @message.subject %></h1>

<div id="content"><%= @message.content %></div>

<%= link_to 'Edit', edit_message_path(@message) %> |
<%= link_to 'Back', messages_path %>
